// tsortbu.cs

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

namespace MergeSort
{
    class Program
    {
        static public void MergeSort(ulong[] a)
        {
            if(a.Length < 2)
                return;
            int n = a.Length;
            ulong [] b = new ulong[n];
            int s = 1;                  // run length
            if((GetPassCount(n) & 1) != 0){     // if would be odd # passes
                for(s = 1; s < n; s += 2){      //   swap in place
                    ulong t;
                    if(a[s] < a[s-1]){
                        t = a[s-1];
                        a[s-1] = a[s];
                        a[s] = t;}
                }
                s = 2;
            }
            for( ; s < n; s *= 2){
                int ee = 0;                     // init end index
                while(ee < n){                  // merge pairs of runs
                    int ll = ee;                // ll = start of left  run
                    int rr = ll+s;              // rr = start of right run
                    if(rr >= n){                // if only left run
                        rr = n;
                        Copy(a, b, ll, rr);     //   copy left run
                        break;                  //   end of pass
                    }
                    ee = rr+s;                  // ee = end of right run
                    if(ee > n)
                        ee = n;
                    Merge(a, b, ll, rr, ee);
                }
                {                               // swap a and b
                    ulong []t = a;
                    a = b;
                    b = t;
                }
            }
        }

        static public void Merge(ulong[] a, ulong[] b, int ll, int rr, int ee)
        {
            int o = ll;                         // b[]       index
            int l = ll;                         // a[] left  index
            int r = rr;                         // a[] right index

            while (true) {                      // merge data
                if (a[l] <= a[r]) {             // if a[l] <= a[r]
                    b[o++] = a[l++];            //   copy a[l]
                    if (l < rr)                 //   if not end of left run
                        continue;               //     continue (back to while)
                    do                          //   else copy rest of right run
                        b[o++] = a[r++];
                    while (r < ee);
                    break;                      //     and return
                } else {                        // else a[l] > a[r]
                    b[o++] = a[r++];            //   copy a[r]
                    if (r < ee)                 //   if not end of right run
                        continue;               //     continue (back to while)
                    do                          //   else copy rest of left run
                        b[o++] = a[l++];
                    while (l < rr);
                    break;                      //     and return
                }
            }
        }

        static public void Copy(ulong[] a, ulong[] b, int ll, int rr)
        {
            while(ll < rr){                     // copy run
                b[ll] = a[ll];
                ll++;
            }
        }

        static public int GetPassCount(int n)   // return number of passes
        {
            int i = 0;
            for(int s = 1; s < n; s <<= 1)
                i += 1;
            return(i);
        }

        // pseudo random ulong linear congruential generator, cycle = 2^64
        static ulong rlcgval = 1;
        const ulong rlcgmpy = 0x5851F42D4C957F2D;   // this value mod 8 == 5
        const ulong rlcgadd = 0x14057B7EF767814F;   // this value mod 2 == 1
        static ulong Rlcg()
        {
            rlcgval = rlcgval * rlcgmpy + rlcgadd;
            return rlcgval;
        }

        static void Main(string[] args)
        {
            const int SIZE = 16*1024*1024;
            ulong [] a = new ulong[SIZE];
            Stopwatch sw = new Stopwatch();
            for (int i = 0; i < a.Length; i++)
                a[i] = Rlcg();
            sw.Start();
            MergeSort(a);
            sw.Stop();
            for (int i = 1; i < a.Length; i++)
            {
                if(a[i] < a[i-1])
                    Console.WriteLine("failed");
            }
            Console.WriteLine("milliseconds: {0:D}\n",sw.ElapsedMilliseconds);
        }
    }
}
