// tsorttd.cs

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

namespace MergeSort
{
    class Program
    {
        static public void MergeSort(ulong[] a)
        {
            if(a.Length < 2)
                return;
            int n = a.Length;
            ulong [] b = new ulong[n];
            TopDownMergeSortAtoA(a, b, 0, a.Length);
        }

        static public void TopDownMergeSortAtoA(ulong[] a, ulong[] b, int ll, int ee)
        {
            if (ee - ll > 1) {
                int rr = (ll + ee)>>1;          // midpoint, start of right half
                TopDownMergeSortAtoB(a, b, ll, rr);
                TopDownMergeSortAtoB(a, b, rr, ee);
                Merge(b, a, ll, rr, ee);        // merge b to a
            }
        }

        static public void TopDownMergeSortAtoB(ulong[] a, ulong[] b, int ll, int ee)
        {
            if (ee - ll > 1) {
                int rr = (ll + ee)>>1;          // midpoint, start of right half
                TopDownMergeSortAtoA(a, b, ll, rr);
                TopDownMergeSortAtoA(a, b, rr, ee);
                Merge(a, b, ll, rr, ee);        // merge a to b
            } else if ((ee - ll) == 1) {
                b[ll] = a[ll];
            }
        }

        static public void Merge(ulong[] a, ulong[] b, int ll, int rr, int ee)
        {
            int o = ll;                         // b[]       index
            int l = ll;                         // a[] left  index
            int r = rr;                         // a[] right index

            while (true) {                      // merge data
                if (a[l] <= a[r]) {             // if a[l] <= a[r]
                    b[o++] = a[l++];            //   copy a[l]
                    if (l < rr)                 //   if not end of left run
                        continue;               //     continue (back to while)
                    do                          //   else copy rest of right run
                        b[o++] = a[r++];
                    while (r < ee);
                    break;                      //     and return
                } else {                        // else a[l] > a[r]
                    b[o++] = a[r++];            //   copy a[r]
                    if (r < ee)                 //   if not end of right run
                        continue;               //     continue (back to while)
                    do                          //   else copy rest of left run
                        b[o++] = a[l++];
                    while (l < rr);
                    break;                      //     and return
                }
            }
        }

        // pseudo random ulong linear congruential generator, cycle = 2^64
        static ulong rlcgval = 1;
        const ulong rlcgmpy = 0x5851F42D4C957F2D;   // this value mod 8 == 5
        const ulong rlcgadd = 0x14057B7EF767814F;   // this value mod 2 == 1
        static ulong Rlcg()
        {
            rlcgval = rlcgval * rlcgmpy + rlcgadd;
            return rlcgval;
        }

        static void Main(string[] args)
        {
            const int SIZE = 16*1024*1024;
            ulong [] a = new ulong[SIZE];
            Stopwatch sw = new Stopwatch();
            for (int i = 0; i < a.Length; i++)
                a[i] = Rlcg();
            sw.Start();
            MergeSort(a);
            sw.Stop();
            for (int i = 1; i < a.Length; i++)
            {
                if(a[i] < a[i-1])
                    Console.WriteLine("failed");
            }
            Console.WriteLine("milliseconds: {0:D}\n",sw.ElapsedMilliseconds);
        }
    }
}
